@@ -1,14 +1,14 @@
-/*	This is pstree written by Fred Hucht (c) 1993-2010	*
- *	EMail: fred AT thp.Uni-Duisburg.de			*
+/*	This is pstree written by Fred Hucht (c) 1993-2013	*
+ *	EMail: fred AT thp.uni-due.de				*
  *	Feel free to copy and redistribute in terms of the	*
  * 	GNU public license. 					*
  *
- * $Id: pstree.c,v 2.33 2009-11-10 22:12:39+01 fred Exp $
+ * $Id: pstree.c,v 2.36 2013-04-12 11:47:03+02 fred Exp $
  */
 static char *WhatString[]= {
-  "@(#)pstree $Revision: 2.33 $ by Fred Hucht (C) 1993-2007",
-  "@(#)EMail: fred AT thp.Uni-Duisburg.de",
-  "$Id: pstree.c,v 2.33 2009-11-10 22:12:39+01 fred Exp $"
+  "@(#)pstree $Revision: 2.36 $ by Fred Hucht (C) 1993-2013",
+  "@(#)EMail: fred AT thp.uni-due.de",
+  "$Id: pstree.c,v 2.36 2013-04-12 11:47:03+02 fred Exp $"
 };
 
 #define MAXLINE 8192
@@ -389,35 +389,43 @@ int GetProcessesDirect(void) {
   }
   
   for (i = j = 0; i < globbuf.gl_pathc; i++) {
-    char name[32];
+    char *pdir, name[32];
     int c;
     FILE *tn;
-    struct stat stat;
     int k = 0;
     
+    pdir = globbuf.gl_pathv[globbuf.gl_pathc - i - 1];
+    
+    /* if processes change their UID this change is only reflected in the owner of pdir.
+     * fixed since version 2.36 */
+    {
+      struct stat st;
+      if (stat(pdir, &st) != 0) { /* get uid */
+	continue; /* process vanished since glob() */
+      }
+      P[j].uid = st.st_uid;
+      uid2user(P[j].uid, P[j].name, sizeof(P[j].name));
+    }
+    
     snprintf(name, sizeof(name), "%s%s",
 	     globbuf.gl_pathv[globbuf.gl_pathc - i - 1], "/stat");
     tn = fopen(name, "r");
     if (tn == NULL) continue; /* process vanished since glob() */
     fscanf(tn, "%ld %s %*c %ld %ld",
 	   &P[j].pid, P[j].cmd, &P[j].ppid, &P[j].pgid);
-    fstat(fileno(tn), &stat);
-    P[j].uid = stat.st_uid;
     fclose(tn);
     P[j].thcount = 1;
     
     snprintf(name, sizeof(name), "%s%s",
 	     globbuf.gl_pathv[globbuf.gl_pathc - i - 1], "/cmdline");
     tn = fopen(name, "r");
-    if (tn == NULL) continue;
+    if (tn == NULL) continue; /* process vanished since glob() */
     while (k < MAXLINE - 1 && EOF != (c = fgetc(tn))) {
       P[j].cmd[k++] = c == '\0' ? ' ' : c;
     }
     if (k > 0) P[j].cmd[k] = '\0';
     fclose(tn);
     
-    uid2user(P[j].uid, P[j].name, sizeof(P[j].name));
-    
 #ifdef DEBUG
     if (debug) fprintf(stderr,
 		       "uid=%5ld, name=%8s, pid=%5ld, ppid=%5ld, pgid=%5ld, thcount=%ld, cmd='%s'\n",
@@ -920,24 +928,66 @@ static char * strstr(s1, s2)
 #endif /* NEED_STRSTR */
 
 #ifdef NEED_SNPRINTF
-/* Contributed by Michael E White */
-int snprintf(char *name, int namesiz, char *format, ...)
-{
-  int retval;
-  char bigbuf[1024] = {'\0'};  
-  va_list ap;
-  va_start(ap, format);
-  retval = vsprintf(bigbuf,format,ap);
-  va_end(ap);
-  if (retval > namesiz) retval = namesiz;
-  strncpy(name, bigbuf, retval);
-  name[retval] = '\0';
-  return retval;
+int snprintf (char *s, int namesiz, char *format, ...) {      
+  /* Original portable version by Michael E. White.
+     This version of Stan Sieler (sieler AT allegro.com) */
+
+  int  chars_needed;              /* not including trailing null */
+  
+  char bigbuf [1024] = {'\0'};    /* note: 1024 is a guess, and may not be large enough! */
+  
+  va_list ap;         /* some systems allow "va_list ap = NULL;", others *do not* (like MACH) */
+  
+  va_start (ap, format);
+  chars_needed = vsprintf (bigbuf, format, ap); /* note: chars_needed does not include trailing null */
+  va_end (ap);
+
+  /* 0 is documented as "don't write anything" ... while not specifically spelled out
+     (e.g., does it also mean "don't internally call vsprintf"?), one can imply that it simply means
+     "don't write to the output buffer 's'.  (Otherwise, if we didn't call vsprintf, we wouldn't
+     know what value of chars_needed to return!) */
+
+   if (namesiz <= 0)
+     ;     /* Don't touch 's' buffer at all! Note: on some systems, a negative namesiz
+	      will cause the process to abort. By checking for <= 0, not just 0, we differ
+	      in that area, but it's a reasonable difference. */
+   
+   else if (chars_needed >= namesiz)  
+     {     /* oh oh, output too large for 'name' buffer... */
+       memcpy (s, bigbuf, namesiz - 1);
+       s [namesiz - 1] = '\0';
+     }
+   
+   else    /* size is ok */
+     {
+       memcpy (s, bigbuf, chars_needed); /* chars_needed < namesiz */
+       s [chars_needed] = '\0';
+       /* note: above two could be replaced by strcpy (s, bigbuf)
+	  since we know strlen (bigbuf) is acceptable.  
+	  But, why copy byte at a time, comparing to null, when
+	  we *know* the length? */
+     }
+   
+   return chars_needed;    /* May be larger than namesiz, but that's ok
+			      In fact, not just 'ok', it's *useful*! */
 }
 #endif  /* NEED_SNPRINTF */
 
 /*
  * $Log: pstree.c,v $
+ * Revision 2.36  2013-04-12 11:47:03+02  fred
+ * Some processes like apache under a recent Linux were listed with UID
+ * root instead of the correct UID, as they use setuid(). We now read the
+ * UID from the owner of /proc/PID instead of /proc/PID/stat, as this
+ * seems to be updated correctly. Thanks to Tom Schmidt
+ * <tschmidt AT micron.com> for pointing out this bug.
+ *
+ * Revision 2.35  2013-02-28 08:33:02+01  fred
+ * Added Stan Sieler's fix to my adaption of snprintf fix by Stan Sieler :-)
+ *
+ * Revision 2.34  2013-02-27 16:57:25+01  fred
+ * Added snprintf fix by Stan Sieler
+ *
  * Revision 2.33  2009-11-10 22:12:39+01  fred
  * Added UTF8, enlarged MAXLINE
  *
